@@ -45,6 +45,7 @@ function _knn(tree::BruteTree{V},
                  point::AbstractVector,
                  best_idxs::Union{Integer, AbstractVector{<:Integer}},
                  best_dists::Union{Number, AbstractVector},
+                 ::Union{Nothing, AbstractVector},
                  skip::F) where {V, F}
 
     return knn_kernel!(tree, point, best_idxs, best_dists, skip, nothing)
@@ -108,3 +109,19 @@ function inrange_kernel!(tree::BruteTree,
     end
     return count
 end
+
+function _inrange_pairs(tree::BruteTree{V}, radius::Number, sortres, skip::F) where {V, F}
+    pairs = NTuple{2,Int}[]
+    for i in 1:length(tree.data)
+        skip(i) && continue
+        for j in (i + 1):length(tree.data)
+            skip(j) && continue
+            d = evaluate(tree.metric, tree.data[i], tree.data[j])
+            if d <= radius
+                push!(pairs, (i, j))
+            end
+        end
+    end
+    sortres && sort!(pairs)
+    return pairs
+end
