@@ -31,11 +31,13 @@ const trees_with_brute = [BruteTree; trees]
 
 include("test_knn.jl")
 include("test_inrange.jl")
+include("test_tree_tree_inrange.jl")
 include("test_monkey.jl")
 include("test_datafreetree.jl")
 include("test_tree_data.jl")
 include("test_periodic.jl")
 include("test_tree_nan.jl")
+include("test_unitful.jl")
 
 @testset "views of SVector" begin
     x = [rand(SVector{3}) for i in 1:20]
